#include <LPC213x.H>                       /* LPC213x definitions  */
#include "types.h"
#include "delay.h"
#include "spi.h"
#include "mmc.h"
#include "fat16.h"


#define MMC_CS_DDR	IODIR0
#define MMC_CS_SET	IOSET0
#define MMC_CS_CLR	IOCLR0
#define PIN_MMC_CS	(1<<20) //Pin 0.20


#ifdef SPI_POLLING
#define spi_send_async spi_io
#define spi_wait_completion() 
#endif

// look in datasheet for complete list of commands
#define MMC_CMD_0_GO_IDLE				0
#define MMC_CMD_1_SEND_OP_COND			1
#define MMC_CMD_9_SEND_CSD				9
#define MMC_CMD_10_SEND_CID				10
#define MMC_CMD_12_STOP					12
#define MMC_CMD_13_SEND_STATUS			13
#define MMC_CMD_16_BLOCKLEN				16
#define MMC_CMD_17_READ_SINGLE			17
#define MMC_CMD_18_READ_MULTIPLE		18
#define MMC_CMD_24_WRITE_SINGLE			0x18

// bit definitions for R1
#define MMC_R1_IN_IDLE 	0x01
#define MMC_R1B_BUSY_BYTE 0x00

// different defines for tokens etc ...
#define MMC_START_TOKEN_SINGLE  0xfe
#define MMC_START_TOKEN_MULTI   0xfc


// starts the read process of a sector
static void MMC_get_sec_start(u32up sector);

// gets the next byte from the MMC card
static u08up MMC_get_sec_next(void);
// stop read process of a sector
static void MMC_get_sec_stop(void);

// deselects the MMC
__inline void MMC_CS_deselect(void);

// selects the MMC for transmission
__inline void MMC_CS_select(void);

static void MMC_cleanup(void);


//---------------------------------------


u32 last_sector;
static u16 current_blocklen = 0;
u08 mmc_open = 0;

// sends a block of data from the mem over the spi bus
static void spi_io_mem(u08* data, u16up length) {
	// transmit 'length' bytes over spi
	while (length) {
		spi_io(*data);
		data++;
		length--;
	}
}

// send cmd + arguments + default crc for init command. once in spi mode
// we dont need crc so we can keep this constant
static void MMC_send_cmd(u08up cmd, u32up data) {
	// default command sequence
	u08 buffer[6];
	// fill sequence with our specific data
	buffer[0] = 0x40 | cmd;
	buffer[1] = data>>24;
	buffer[2] = data>>16;
	buffer[3] = data>>8;
	buffer[4] = data;
	buffer[5] = 0x95;
	// dispatch data
	spi_io_mem(buffer, 6);
}

// gets a 1 byte long R1
static u08up MMC_get_R1(void) {
	u08up retval;
	u08up max_errors = 128;
	// wait for first valid response byte
	do {
		retval = spi_io(0xff);
		max_errors--;
	}
	while((retval & 0x80) && max_errors);
	return retval;
}

// gets a 2 byte long R2
/*static u16up MMC_get_R2(void) {
	u16up retval;
	u08up max_errors = 64;
	// wait for first valid response byte
	do {
		retval = spi_io(0xff);
		max_errors--;
	}
	while((retval & 0x80) && max_errors);
	// move data to upper byte
	retval = (retval << 8);
	// get second byte;
	max_errors = spi_io(0xff);
	retval += max_errors;
	return retval;
}*/

// sends the start data block token to the MMC
/*static void MMC_send_start_data_token(void) {
  spi_io(MMC_START_TOKEN_SINGLE);
}*/

// waits for the card to send the start data block token
static void MMC_wait_for_start_token(void) {
  u08up retval;
  do
    retval = spi_io(0xff);
  while ((retval != MMC_START_TOKEN_SINGLE));
}

// selects the CS for spi xfer
__inline void MMC_CS_select(void) {
	// pull down the MMC CS line
	MMC_CS_CLR = PIN_MMC_CS;
}

// deselects the CS for spi xfer
__inline void MMC_CS_deselect(void) {
	// pull up the MMC CS card
	MMC_CS_SET = PIN_MMC_CS;
}


// stops the MMC transmission and sends the clock cycles needed by the mmc for cleanup
static void MMC_cleanup(void) {
	// deselect the MMC card
	MMC_CS_deselect();
	// pulse the SCK 16 times
	spi_io(0xff);
	spi_io(0xff);
}

// gets n bytes plus crc from spi bus
static void MMC_get_data(u08* ptr_data, u16up length) {
  MMC_wait_for_start_token();
  while (length) {
    *ptr_data = spi_io(0xff);
	length--;
    ptr_data++;
  }
  // get the 2 CRC bytes
  spi_io(0xff);
  spi_io(0xff);
}


// reads the CID reg from the card
/*static void MMC_get_CID(u08 *ptr_data) {
  // select card
	MMC_CS_select();
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_10_SEND_CID, 0);
	// get the response
	MMC_get_R1();
	// get the register data
	MMC_get_data(ptr_data, 16);
	// cleanup behind us
	MMC_cleanup();
}*/



// reads the CSD reg from the card
static void MMC_get_CSD(u08 *ptr_data) {
	// select card
	MMC_CS_select();
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_9_SEND_CSD, 0);
	// get the response
	MMC_get_R1();
	// get the register data
	MMC_get_data(ptr_data, 16);
	// cleanup behind us
	MMC_cleanup();
}

// set the BLOCKLEN for transmissions
static void MMC_set_blocklen(u16up blocklen) {
	// make sure this block len is not already set
	if (current_blocklen != blocklen) {
		current_blocklen = blocklen;
		// select card
		MMC_CS_select();
		// tell the MMC card that we want to know its status
		MMC_send_cmd(MMC_CMD_16_BLOCKLEN, blocklen);
		// get the response
		MMC_get_R1();
		// cleanup behind us
		MMC_cleanup();
	}
}



// returns the :
// 		size of the card in MB ( ret * 1024^2) == bytes
// 		sector count and multiplier MB are in u08 == C_SIZE / (2^(9-C_SIZE_MULT))
// 		name of the media
static void MMC_get_volume_info(void) {//VOLUME_INFO* vinf){
	u08 data[16];
	// read the CSD register
	MMC_get_CSD(data);
	// get the C_SIZE value. bits [73:62] of data
	// [73:72] == data[6] && 0x03
	// [71:64] == data[7]
	// [63:62] == data[8] && 0xc0
	last_sector = data[6] & 0x03;
	last_sector <<= 8;
	last_sector += data[7];
	last_sector <<= 2;
	last_sector += (data[8] & 0xc0) >> 6;
}

// sets up the pins used by the MMC interface
void MMC_hw_init(void) {
	// set MMC CS pin high output
	MMC_CS_DDR |= PIN_MMC_CS;
	MMC_CS_SET = PIN_MMC_CS;
}

// starts up the MMC card
u08up MMC_init(void) {
	u08up i;
	u08up res;
	// the data sheet says that the MMC needs 74 clock pulses to startup
	// 10*8== 80; 80>76
	for( i = 0; i < 10; i++)
		spi_io(0xff);

	// select card
	MMC_CS_select();
	// put MMC in idle
	MMC_send_cmd(MMC_CMD_0_GO_IDLE, 0);
	// get the response
	res = MMC_get_R1();

	//if(res != 0x01) USART_sendstring("Error R1-init",0x01);
	if (res != 0x01)
		return 0;

	// tell the MMC to start its init process by sending the MMC_CMD_1_SEND_OP_COND comand
	// until the response has the idle bit set to 0
	while (res == 0x01) {
		// deselect card
		MMC_CS_deselect();
		// send 32 clock pulses
		spi_io(0xff);
		spi_io(0xff);
		spi_io(0xff);
		spi_io(0xff);
		// select card
		MMC_CS_select();
    // send wake up signal s.t. MMC leaves idle state and switches to operation mode
		MMC_send_cmd(MMC_CMD_1_SEND_OP_COND, 0);
    // get response
		res = MMC_get_R1();
	}
	// cleanup behind us
	MMC_cleanup();

	// now increase MMC card interface speed
	SSPCR0 = 0x0207; // 7=3,75MHz SPI

	// find out some info on card
	MMC_get_volume_info();

	MMC_set_blocklen(512);
	mmc_open = 0;
	return 1;
}

// starts the read process of a sector
static void MMC_get_sec_start(u32 sector) {
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sector <<= 9;
	// select card
	MMC_CS_select();
	mmc_open = 1;
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_17_READ_SINGLE, sector);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// wait till the mmc starts sending data
	MMC_wait_for_start_token();
}

// gets the next byte from the MMC card
__inline u08up MMC_get_sec_next(void) {
	u08up tmp = spi_io(0xff);
	return tmp;
}

// stop read process of a sector
static void MMC_get_sec_stop(void) {
	if (mmc_open) {
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
		// give enough time
		MMC_cleanup();
		mmc_open = 0;
	}
}


// starts the write process of a sector
static void MMC_write_sec_start(u32up sector) {
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sector <<= 9;
	// select card
	MMC_CS_select();
	mmc_open = 1;
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_24_WRITE_SINGLE, sector);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// send start byte
	spi_io(0xfe);
}

// writes the next byte to the MMC card
__inline void MMC_write_sec_next(u08up b) {
	spi_io(b);
}

// stop write process of a sector
static void MMC_get_sec_stop_w(void) {
	if (mmc_open) {
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
	    // wait while card is busy
	    while (spi_io(0xff) != 0xff);
		// give enough time
		MMC_cleanup();
		mmc_open = 0;
	}
}



//-------------------------------------------------------------
// High Level MMC-Funktionen

u32 sect;
u16 bufpos=0;

extern u08 fat_directaccess;

void mmc_read_sector(u32up s, u08 *buf) {
	u16up i;
	MMC_get_sec_start(s);
	for (i=0; i<512; i++)
		buf[i] = MMC_get_sec_next();
	MMC_get_sec_stop();
}

// Lesen beginnen
void mmc_load_start(u32up s) {
	sect = s;
	MMC_get_sec_start(s);
	bufpos = 0;
}


// Byte lesen
u08up mmc_fetch_byte(void) {
	if (bufpos == 512) {
		MMC_get_sec_stop();
		if (!fat_directaccess)
			fat_getnextsector();
		else
			sect++;
		mmc_load_start(sect);
	}
	bufpos++;
	return MMC_get_sec_next();
}


// Lesen, um die 512 voll zu machen
void mmc_complete_read(void) {
	while (bufpos < 512) {
		MMC_get_sec_next();
		bufpos++;
	}
	MMC_get_sec_stop();
}


void mmc_read_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		*buf = mmc_fetch_byte();
		buf++;
	}
}


void mmc_write_sector(u32up s, u08 *buf) {
	u16up i;
	MMC_write_sec_start(s);
	for (i=0; i<512; i++)
		spi_send_async(buf[i]);
	spi_wait_completion();
	MMC_get_sec_stop_w();
}

// Schreiben beginnen
void mmc_write_start(u32up s) {
	sect = s;
#ifndef BUFFERED_SECTOR
	MMC_write_sec_start(s);
#endif
	bufpos = 0;
}


// Byte schreiben
#ifdef BUFFERED_SECTOR
u08 sd_buf[512];
void mmc_write_byte(u08up b) {
	u32up i;
	sd_buf[bufpos++] = b;
	if (bufpos == 512) {
		bufpos = 0;
		MMC_write_sec_start(sect);
		for (i=0; i<512; i++)
			spi_send_async(sd_buf[i]);
		spi_wait_completion();
		MMC_get_sec_stop_w();
		if (!fat_directaccess)
			fat_getnextsector();
		else
			sect++;
	}
}
void mmc_complete_write(void) {
	u32up i;
	if (bufpos == 0) return;
	while (bufpos < 512)
		sd_buf[bufpos++] = 0;
	MMC_write_sec_start(sect);
	for (i=0; i<512; i++)
		spi_send_async(sd_buf[i]);
	spi_wait_completion();
	MMC_get_sec_stop_w();
}
#else
void mmc_write_byte(u08up b) {
	if (bufpos == 512) {
		spi_wait_completion();
		MMC_get_sec_stop_w();
		if (!fat_directaccess)
			fat_getnextsector();
		else
			sect++;
		mmc_write_start(sect);
	}
	bufpos++;
	spi_send_async(b&0xff);
}
// Schreiben, um die 512 voll zu machen
void mmc_complete_write(void) {
	while (bufpos < 512) {
		spi_send_async(0);
		bufpos++;
	}
	spi_wait_completion();
	MMC_get_sec_stop_w();
}
#endif



/*void simul_spi(void) {
	u32up i;
	for (i=0;i<512;i++)
		spi_send_async(0xAC);
	spi_wait_completion();
}*/


// Little Endian Write WORD
void mmc_write_word(u16up b) {
	mmc_write_byte(b);
	mmc_write_byte(b>>8);
}

// Little Endian Write DWORD
void mmc_write_dword(u32up b) {
	mmc_write_byte(b);
	mmc_write_byte(b>>8);
	mmc_write_byte(b>>16);
	mmc_write_byte(b>>24);
}


void mmc_write_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		mmc_write_byte(*buf);
		buf++;
	}
}


// Daten vergleichen
u08up mmc_compare_string(u08 *buf, u08up len) {
	u08up i;
	for (i=0; i<len; i++) {
		if (mmc_fetch_byte() != buf[i])
			return 0;
	}
	return 1;
}

